from .new_distributor import new_distributor, NewDistributorArgs, NewDistributorAccounts
from .new_claim import new_claim, NewClaimArgs, NewClaimAccounts
from .claim_locked import claim_locked, ClaimLockedAccounts
from .clawback import clawback, ClawbackAccounts
from .set_admin import set_admin, SetAdminAccounts
from .withdraw_custody_sol import (
    withdraw_custody_sol,
    WithdrawCustodySolArgs,
    WithdrawCustodySolAccounts,
)



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from spl.token.constants import TOKEN_PROGRAM_ID, ASSOCIATED_TOKEN_PROGRAM_ID
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class NewDistributorArgs(typing.TypedDict):
    version: int
    root: list[int]
    max_total_claim: int
    max_num_nodes: int
    start_vesting_ts: int
    end_vesting_ts: int


layout = borsh.CStruct(
    "version" / borsh.U64,
    "root" / borsh.U8[32],
    "max_total_claim" / borsh.U64,
    "max_num_nodes" / borsh.U64,
    "start_vesting_ts" / borsh.I64,
    "end_vesting_ts" / borsh.I64,
)


class NewDistributorAccounts(typing.TypedDict):
    distributor: Pubkey
    mint: Pubkey
    token_vault: Pubkey
    creator: Pubkey


def new_distributor(
    args: NewDistributorArgs,
    accounts: NewDistributorAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["distributor"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["token_vault"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["creator"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=ASSOCIATED_TOKEN_PROGRAM_ID, is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b" \x8bp\xab\x00\x02\xe1\x9b"
    encoded_args = layout.build(
        {
            "version": args["version"],
            "root": args["root"],
            "max_total_claim": args["max_total_claim"],
            "max_num_nodes": args["max_num_nodes"],
            "start_vesting_ts": args["start_vesting_ts"],
            "end_vesting_ts": args["end_vesting_ts"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



import typing
from dataclasses import dataclass
from solders.pubkey import Pubkey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID


class MerkleDistributorJSON(typing.TypedDict):
    bump: int
    version: int
    root: list[int]
    mint: str
    token_vault: str
    max_total_claim: int
    max_num_nodes: int
    total_amount_claimed: int
    num_nodes_claimed: int
    start_ts: int
    end_ts: int
    creator: str
    admin: str
    clawed_back: bool


@dataclass
class MerkleDistributor:
    discriminator: typing.ClassVar = b"Mw\x8bFT\xf7\x0c\x1a"
    layout: typing.ClassVar = borsh.CStruct(
        "bump" / borsh.U8,
        "version" / borsh.U64,
        "root" / borsh.U8[32],
        "mint" / BorshPubkey,
        "token_vault" / BorshPubkey,
        "max_total_claim" / borsh.U64,
        "max_num_nodes" / borsh.U64,
        "total_amount_claimed" / borsh.U64,
        "num_nodes_claimed" / borsh.U64,
        "start_ts" / borsh.I64,
        "end_ts" / borsh.I64,
        "creator" / BorshPubkey,
        "admin" / BorshPubkey,
        "clawed_back" / borsh.Bool,
    )
    bump: int
    version: int
    root: list[int]
    mint: Pubkey
    token_vault: Pubkey
    max_total_claim: int
    max_num_nodes: int
    total_amount_claimed: int
    num_nodes_claimed: int
    start_ts: int
    end_ts: int
    creator: Pubkey
    admin: Pubkey
    clawed_back: bool

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["MerkleDistributor"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["MerkleDistributor"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["MerkleDistributor"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "MerkleDistributor":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = MerkleDistributor.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            bump=dec.bump,
            version=dec.version,
            root=dec.root,
            mint=dec.mint,
            token_vault=dec.token_vault,
            max_total_claim=dec.max_total_claim,
            max_num_nodes=dec.max_num_nodes,
            total_amount_claimed=dec.total_amount_claimed,
            num_nodes_claimed=dec.num_nodes_claimed,
            start_ts=dec.start_ts,
            end_ts=dec.end_ts,
            creator=dec.creator,
            admin=dec.admin,
            clawed_back=dec.clawed_back,
        )

    def to_json(self) -> MerkleDistributorJSON:
        return {
            "bump": self.bump,
            "version": self.version,
            "root": self.root,
            "mint": str(self.mint),
            "token_vault": str(self.token_vault),
            "max_total_claim": self.max_total_claim,
            "max_num_nodes": self.max_num_nodes,
            "total_amount_claimed": self.total_amount_claimed,
            "num_nodes_claimed": self.num_nodes_claimed,
            "start_ts": self.start_ts,
            "end_ts": self.end_ts,
            "creator": str(self.creator),
            "admin": str(self.admin),
            "clawed_back": self.clawed_back,
        }

    @classmethod
    def from_json(cls, obj: MerkleDistributorJSON) -> "MerkleDistributor":
        return cls(
            bump=obj["bump"],
            version=obj["version"],
            root=obj["root"],
            mint=Pubkey.from_string(obj["mint"]),
            token_vault=Pubkey.from_string(obj["token_vault"]),
            max_total_claim=obj["max_total_claim"],
            max_num_nodes=obj["max_num_nodes"],
            total_amount_claimed=obj["total_amount_claimed"],
            num_nodes_claimed=obj["num_nodes_claimed"],
            start_ts=obj["start_ts"],
            end_ts=obj["end_ts"],
            creator=Pubkey.from_string(obj["creator"]),
            admin=Pubkey.from_string(obj["admin"]),
            clawed_back=obj["clawed_back"],
        )



import typing
from anchorpy.error import ProgramError


class InsufficientUnlockedTokens(ProgramError):
    def __init__(self) -> None:
        super().__init__(6000, "Insufficient unlocked tokens")

    code = 6000
    name = "InsufficientUnlockedTokens"
    msg = "Insufficient unlocked tokens"


class InvalidProof(ProgramError):
    def __init__(self) -> None:
        super().__init__(6001, "Invalid Merkle proof.")

    code = 6001
    name = "InvalidProof"
    msg = "Invalid Merkle proof."


class ExceededMaxClaim(ProgramError):
    def __init__(self) -> None:
        super().__init__(6002, "Exceeded maximum claim amount")

    code = 6002
    name = "ExceededMaxClaim"
    msg = "Exceeded maximum claim amount"


class MaxNodesExceeded(ProgramError):
    def __init__(self) -> None:
        super().__init__(6003, "Exceeded maximum node count")

    code = 6003
    name = "MaxNodesExceeded"
    msg = "Exceeded maximum node count"


class Unauthorized(ProgramError):
    def __init__(self) -> None:
        super().__init__(6004, "Account is not authorized to execute this instruction")

    code = 6004
    name = "Unauthorized"
    msg = "Account is not authorized to execute this instruction"


class OwnerMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(6005, "Token account owner did not match intended owner")

    code = 6005
    name = "OwnerMismatch"
    msg = "Token account owner did not match intended owner"


class ClawbackAlreadyClaimed(ProgramError):
    def __init__(self) -> None:
        super().__init__(6006, "Clawback already claimed")

    code = 6006
    name = "ClawbackAlreadyClaimed"
    msg = "Clawback already claimed"


class SameAdmin(ProgramError):
    def __init__(self) -> None:
        super().__init__(6007, "New and old admin are identical")

    code = 6007
    name = "SameAdmin"
    msg = "New and old admin are identical"


class ClaimExpired(ProgramError):
    def __init__(self) -> None:
        super().__init__(6008, "Claim window expired")

    code = 6008
    name = "ClaimExpired"
    msg = "Claim window expired"


class ArithmeticError(ProgramError):
    def __init__(self) -> None:
        super().__init__(6009, "Arithmetic Error (overflow/underflow)")

    code = 6009
    name = "ArithmeticError"
    msg = "Arithmetic Error (overflow/underflow)"


class StartTimestampAfterEnd(ProgramError):
    def __init__(self) -> None:
        super().__init__(6010, "Start Timestamp cannot be after end Timestamp")

    code = 6010
    name = "StartTimestampAfterEnd"
    msg = "Start Timestamp cannot be after end Timestamp"


class TimestampsNotInFuture(ProgramError):
    def __init__(self) -> None:
        super().__init__(6011, "Timestamps cannot be in the past")

    code = 6011
    name = "TimestampsNotInFuture"
    msg = "Timestamps cannot be in the past"


class InsufficientFunds(ProgramError):
    def __init__(self) -> None:
        super().__init__(6012, "Insufficient funds")

    code = 6012
    name = "InsufficientFunds"
    msg = "Insufficient funds"


CustomError = typing.Union[
    InsufficientUnlockedTokens,
    InvalidProof,
    ExceededMaxClaim,
    MaxNodesExceeded,
    Unauthorized,
    OwnerMismatch,
    ClawbackAlreadyClaimed,
    SameAdmin,
    ClaimExpired,
    ArithmeticError,
    StartTimestampAfterEnd,
    TimestampsNotInFuture,
    InsufficientFunds,
]
CUSTOM_ERROR_MAP: dict[int, CustomError] = {
    6000: InsufficientUnlockedTokens(),
    6001: InvalidProof(),
    6002: ExceededMaxClaim(),
    6003: MaxNodesExceeded(),
    6004: Unauthorized(),
    6005: OwnerMismatch(),
    6006: ClawbackAlreadyClaimed(),
    6007: SameAdmin(),
    6008: ClaimExpired(),
    6009: ArithmeticError(),
    6010: StartTimestampAfterEnd(),
    6011: TimestampsNotInFuture(),
    6012: InsufficientFunds(),
}


def from_code(code: int) -> typing.Optional[CustomError]:
    maybe_err = CUSTOM_ERROR_MAP.get(code)
    if maybe_err is None:
        return None
    return maybe_err



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from spl.token.constants import TOKEN_PROGRAM_ID
from solders.instruction import Instruction, AccountMeta
from construct import Construct
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class NewClaimArgs(typing.TypedDict):
    amount_unlocked: int
    amount_locked: int
    proof: list[list[int]]


layout = borsh.CStruct(
    "amount_unlocked" / borsh.U64,
    "amount_locked" / borsh.U64,
    "proof" / borsh.Vec(typing.cast(Construct, borsh.U8[32])),
)


class NewClaimAccounts(typing.TypedDict):
    distributor: Pubkey
    claim_status: Pubkey
    from_: Pubkey
    to: Pubkey
    claimant: Pubkey


def new_claim(
    args: NewClaimArgs,
    accounts: NewClaimAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["distributor"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["claim_status"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["from_"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["to"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["claimant"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"N\xb1b{\xd2\x15\xbbS"
    encoded_args = layout.build(
        {
            "amount_unlocked": args["amount_unlocked"],
            "amount_locked": args["amount_locked"],
            "proof": args["proof"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class WithdrawCustodySolArgs(typing.TypedDict):
    amount_lamports: int


layout = borsh.CStruct("amount_lamports" / borsh.U64)


class WithdrawCustodySolAccounts(typing.TypedDict):
    sol_custody: Pubkey
    owner: Pubkey


def withdraw_custody_sol(
    args: WithdrawCustodySolArgs,
    accounts: WithdrawCustodySolAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["sol_custody"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["owner"], is_signer=True, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"v\x86T';|\xd9\x9a"
    encoded_args = layout.build(
        {
            "amount_lamports": args["amount_lamports"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from spl.token.constants import TOKEN_PROGRAM_ID
from solders.instruction import Instruction, AccountMeta
from ..program_id import PROGRAM_ID


class ClaimLockedAccounts(typing.TypedDict):
    distributor: Pubkey
    claim_status: Pubkey
    from_: Pubkey
    to: Pubkey
    claimant: Pubkey


def claim_locked(
    accounts: ClaimLockedAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["distributor"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["claim_status"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["from_"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["to"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["claimant"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b'"\xce\xb5\x17\x0b\xcf\x93Z'
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)
